using CSV

include(joinpath("..", "examples", "repeated_wave_propagation.jl"))
include(joinpath("..", "examples", "analyze_turbulent_channel_model.jl"))
include(joinpath("..", "examples", "analyze_coupling_data.jl"))
include(joinpath("..", "examples", "batch_analysis.jl"))
include(joinpath("..", "examples", "from_dump.jl"))


function plot_batch_scintillation_index(dirname; xcols=:, kwargs...)
    # gather all hdf5 files in the given directory `dirname`
    files = readdir(dirname, join=true)
    xidx = findfirst(contains("xvals.csv"), files)
    csvidx = findfirst(contains("scintillation_indices.csv"), files)
    if isnothing(xidx)
        throw(MissingException("File xvals.csv not found."))
    end
    if isnothing(csvidx)
        throw(MissingException("File scintillation_indices.csv not found."))
    end
    # read files
    xdf = CSV.read(files[xidx], DataFrame)
    df = CSV.read(files[csvidx], DataFrame)
    # gather all swept variables
    xnames = names(xdf)
    xnames_plt = if isone(length(xnames))
        first(xnames)
    else
        xnames[xcols]
    end

    plt = _plot_batch_scintillation_index(df, xnames, xnames_plt; kwargs...)
    display(plt)
    return
end


function _plot_batch_scintillation_index(df, xnames::AbstractVector{T}, xnames_plt::T; xscale="linear", yscale="linear", relative=false, reference="gaussian") where {T<:Union{Symbol,String}}
    # get x-axis
    x = df[!, xnames_plt]
    yref = df[!, reference]
    ye = df[!, "estimate"]

    if relative
        plt = plot.scatter(x=x, y=100 * relrms(ye, yref, dims=()), name="Estimate", mode="markers", marker=Config(color="black"), showlegend=true)

        # format y-axis
        plt.layout.yaxis.title = "Rel. error w.r.t. reference / %"
        plt.layout.yaxis.range = [0, 100]
    else
        plt = plot.scatter(x=x, y=yref, name="Reference ($reference)", mode="markers")
        plt.scatter(x=x, y=ye, name="Estimate", mode="markers", marker=Config(color="black"))
        
        # format y-axis
        plt.layout.yaxis.title = "Scintillation index"
        plt.layout.yaxis.range = [0, 1.1 * maximum.(select(df, Not(xnames)) |> eachcol) |> maximum]
    end

    plt.layout.xaxis.title = string(xnames_plt)
    # set x/y axis to logarithmic
    plt.layout.xaxis.type = xscale
    plt.layout.yaxis.type = yscale

    return plt
end

function _plot_batch_scintillation_index(df, xnames::AbstractVector{T}, xnames_plt::AbstractVector{T}; xscale="linear", yscale="linear", relative=false, reference="gaussian") where {T<:Union{Symbol,String}}

    # we do not support anything but two xnames in this function
    @assert length(xnames_plt) == 2

    # get z-data
    z_df = unstack(df, xnames_plt..., "estimate")
    # convert z-data to Matrix
    z = Matrix(z_df[:, 2:end]) |> transpose

    # detect dependent variables by checking whether the off-diagonals contain only NaNs
    zv = vec(z) |> collect
    offdiag = deleteat!(zv, diagind(z))

    if all(ismissing, offdiag)
        # we consider two dependent variables
        ye = diag(z)
        # gather corresponding reference values
        zref_df = unstack(df, xnames_plt..., reference)
        yref = diag(Matrix(zref_df[:, 2:end]))

        # gather both x-axes
        x1 = string.(z_df[!, xnames_plt[1]])
        x2 = names(z_df)[2:end]

        if relative
            err = 100 * relrms(ye, yref, dims=())
            plt = plot.scatter(x=x1, y=err, name="Estimate", mode="markers", marker=Config(color=err, cmin=0, cmax=100), showlegend=true)
            plt.scatter(x=x2, y=fill(NaN, size(x2)), xaxis="x2", showlegend=false)

            # format y-axis
            plt.layout.yaxis.title = "Rel. error w.r.t. reference / %"
            plt.layout.yaxis.range = [0, 100]
        else
            plt = plot.scatter(x=x1, y=yref, name="Reference ($reference)", mode="markers")
            plt.scatter(x=x1, y=ye, name="Estimate", mode="markers", marker=Config(color="black"))
            plt.scatter(x=x2, y=fill(NaN, size(x2)), xaxis="x2", showlegend=false)

            # format y-axis
            plt.layout.yaxis.title = "Scintillation index"
            plt.layout.yaxis.range = [0, 1.1 * maximum.(select(df, Not(xnames)) |> eachcol) |> maximum]
        end

        # format primary xaxis
        plt.layout.xaxis.title = string(xnames_plt[1])
        # format second x-axis
        plt.layout.xaxis2.position = 1.0
        plt.layout.xaxis2.anchor = "free"
        plt.layout.xaxis2.overlaying = "x"
        plt.layout.xaxis2.side = "top"
        plt.layout.xaxis2.title = string(xnames_plt[2])
        # legend
        plt.layout.legend.orientation = "h"
        plt.layout.legend.xanchor = "center"
        plt.layout.legend.x = 0.5
        plt.layout.legend.y = 1.4

        return plt
    else
        # get x-axis, should be categorical!
        x = string.(z_df[!, xnames_plt[1]])
        # get y-axis, should be categorical!
        y = names(z_df)[2:end]

        if relative
            # get z-data
            zref_df = unstack(df, xnames_plt..., reference)
            # convert z-data to Matrix
            zref = Matrix(zref_df[:, 2:end]) |> transpose

            plt = plot.heatmap(x=x, y=y, z=100 * relrms(z, zref, dims=()), name="Estimate", zmin=0, zmax=100, colorscale="Reds")
            
            plt.layout.title = "Rel. error w.r.t. reference / %"
        else
            plt = plot.heatmap(x=x, y=y, z=z, name="Estimate", zmin=0, zmax=1.1 * maximum.(select(df, Not(xnames)) |> eachcol) |> maximum)
            
            plt.layout.title = "Scintillation index"
        end

        plt.layout.xaxis.title = string(xnames_plt[1])
        plt.layout.yaxis.title = string(xnames_plt[2])

        # set x/y axis to logarithmic if needed
        plt.layout.xaxis.type = xscale
        plt.layout.yaxis.type = yscale

        return plt
    end
end