include(joinpath("..", "src", "backpropagate.jl"))
include(joinpath("..", "src", "estimate_statistics.jl"))
include(joinpath("..", "src", "sampling_requirements.jl"))
include(joinpath("..", "src", "turbulent_channel.jl"))
include(joinpath("..", "src", "monte_carlo.jl"))
include(joinpath("..", "src", "optical_parameters.jl"))

include("from_dump.jl")
include("plot_results.jl")

using HDF5
using H5Zblosc
using DataFrames
using FourierTools
using Optim
import FourierTools: center_extract, center_pos
import Dates: format, now, today

# From Zheng 2016 Opt. Expr.: "The first [fiber] is a standard G.652 SMF with fundamental mode diameter of 9 μm. The second one is a 2MF with core diameter of 18 μm and fundamental mode diameter of 15.6 μm. The last one is a 4MF whose core diameter is 24 μm and fundamental mode diameter is 18.2 μm.
# see specs of Corning® SMF-28® Ultra Optical Fiber: https://www.corning.com/media/worldwide/coc/documents/Fiber/product-information-sheets/PI-1424-AEN.pdf
default_fibers = (
    # SMF=(4.1e-6, 0.1, 5.2e-6),
    # FMF3=(9e-6, 0.1, 7.8e-6),
    FMF6=(12e-6, 0.1, 9.1e-6),
)
default_proteus = (; (Symbol("CP$k") => (300e-6, k) for k ∈ 1:6)...)


"""
    fn: base filename of HDF5 file(s) to load data from

"""
function coupling_statistics(templatefn; f=80e-3, wg=FiberModes(default_fibers.FMF6...), basedir=pwd(), showplot=false, dump=true, optim_focus=true, pdf=false, plotmodes=false, samplingcheck=false, showprogress=true)

    # obtain relevant files
    allfiles = readdir(basedir, join=true)
    basefn = templatefn[1:findlast('_', templatefn)]
    hdf5files = filter!(contains(basefn * r"\d+\.hdf5"), allfiles)
    # just in case, sort?
    sort!(hdf5files)

    # load metadata
    metadata = h5open(first(hdf5files), "r") do h5
        Dict(attrs(h5))
    end

    # verify that we have all the chunks at hand
    (metadata["nchunks"] != length(hdf5files)) && @warn "Found $(length(hdf5files)) files, expected $(metadata["nchunks"])"

    dx = metadata["dx"]
    dy = metadata["dy"]
    dz = metadata["dz"]
    # build a domain struct corresponding to receiver plane
    T = eltype(dx)
    domain = Domain{T}(dx[end], dy[end], dz, metadata["Nx"], metadata["Ny"], metadata["Nz"])
    # dummy beam
    beam = Beam(0, 0, metadata["lambda"], metadata["w0"])

    optim_focus && (f = optimize_focus(domain, beam, wg))

    dxf = focal_plane_res(domain, beam, wg, f)[1]
    # sampling requirements satisfied for backpropagation?
    samplingcheck && check_sampling(f, domain.Nx, domain.dx, domain.Dx, dxf, 2wg.a; λ=1.55e-6, R=-f, r0=Inf, c=2, debug=false)

    modes = setup_modes(wg, domain, beam, f)
    for (k, mode) ∈ modes
        backprop!(mode, dxf, domain.dx, f, beam.λ)
    end

    # optionally, plot backpropagated fiber modes
    if plotmodes
        x = fftshift(fftfreq(domain.Nx, domain.Dx))
        y = fftshift(fftfreq(domain.Ny, domain.Dy))

        for (k, mode) ∈ modes
            z = fftshift(Adapt.adapt(Array, mode))
            tr = complex_plot(; x, y, z)
            lt = Config(; title=string(k...), xaxis=Config(title="x"), yaxis=Config(title="y", scaleanchor="x", autorange=true))
            Plot(tr, lt) |> display
        end
    end

    # generate coupling efficiency DataFrame
    η = DataFrame(Dict(string(k) => zeros(metadata["iters"]) for k ∈ keys(modes)))

    # fill coupling efficiency DataFrame
    for (j, chunk) ∈ enumerate(Iterators.partition(1:metadata["iters"], metadata["chunksize"]))

        urx = maybe_cuda(h5read(hdf5files[j], "RXFields"))

        prg = Progress(length(chunk), desc="Chunk $j/$(metadata["nchunks"]): "; enabled=showprogress)

        for (ci, i) ∈ enumerate(chunk)
            for (k, mode) ∈ modes
                @views η[!, string(k)][i] = coupling_efficiency(urx[:, :, ci], mode)
            end

            next!(prg)
        end
    end

    stats = pdf ? estimate_pdf : estimate_ccdf

    if showplot
        traces = Config[]

        for k ∈ keys(modes)
            x, y = stats(-10log10.(η[!, string(k)]))
            push!(traces, Config(type="scatter", x=x, y=y, name=string(k...), showlegend=true))
        end
        Llim = 10(fld(max((maximum(t.x) for t ∈ traces)...), 10) + 1)

        p = Plot(traces, Config(xaxis=Config(title="Coupling loss budget L [dB]", range=[0, Llim]), yaxis=Config(title="Outage probability", type="log", nticks=round(Int, log10(metadata["iters"])) + 2), title="Focal length f=$(round(f*1e3, sigdigits=3)) mm"))

        pdf && (p.layout.yaxis.type = "linear")

        display(p)
    end


    if dump
        # dump to .hdf5 file for postprocessing
        fn = basefn * "coupling.hdf5"

        h5open(fn, "w") do fid

            ηgrp = create_group(fid, "Efficiencies")
            [ηgrp["M"*string(k...)] = v for (k, v) ∈ zip(names(η), eachcol(η))]

            # save mode fields, too -> does not hurt
            mgrp = create_group(fid, "Modefields")
            [mgrp["M"*string(k...)] = Adapt.adapt(Array, v) for (k, v) ∈ pairs(modes)]

            attrs(fid)["focallength"] = f
            [attrs(fid)[k] = v for (k, v) ∈ pairs(metadata)]
        end
    end

    return nothing
end

function scintillation_index(fn; basedir=pwd(), debug=false, showplot=false)

    fn = joinpath(basedir, fn)

    # load data & metadata
    urx, metadata = h5open(fn, "r") do h5
        maybe_cuda(read(h5["RXFields"])), Dict(attrs(h5))
    end

    # coordinate axes
    Nx = metadata["Nx"]
    Ny = metadata["Ny"]
    x = fftfreq(Nx, last(metadata["Dx"]))
    y = fftfreq(Ny, last(metadata["Dy"]))
    # other parameters
    λ = metadata["lambda"]
    Dz = metadata["Dz"]
    Cn² = metadata["Cn2"]
    w0 = metadata["w0"]
    l0 = metadata["l0"]
    L0 = metadata["L0"]

    # scintillation index along highest dimension, note that the dumped fields are not fftshifted!
    SI = squeeze(scv(urx; dims=ndims(urx)))
    SI0 = CUDA.@allowscalar first(SI) # without the fftshift, the first element is the on-axis one

    if debug
        # compute theoretical reference values
        SI_pw = scintillation_index_pw(2π / λ, Dz, Cn², l0, L0)
        SI_sw = scintillation_index_sw(2π / λ, Dz, Cn², l0, L0)
        SI_gaussian = scintillation_index_gaussian(2π / λ, Dz, Cn², w0, l0, L0)

        @info "Estimated on-axis scintillation index: " SI0
        @info "Theoretical result (plane wave): " SI_pw
        @info "Theoretical result (spherical wave): " SI_sw
        @info "Theoretical result (gaussian beam): " SI_gaussian

        # check convergence on axis
        SI_evo = squeeze(cumscv(urx[1, 1, :]))
        # horizontal axis
        x = 1:length(SI_evo)
        # estimation std based on delta method (Gemini)
        estim_std = @. SI_evo * sqrt((1 + SI_evo^2) / (2x))
        # convert to CPU arrays if necessary for plotting
        y = Adapt.adapt(Array, SI_evo)
        yerr = Adapt.adapt(Array, estim_std)
        # plot
        p = plot.scatter(x=x, y=y, name="Evolution of estimated SI")
        p.scatter(x=vcat(x, reverse(x)), y=vcat(y .+ 3 .* yerr, reverse(y .- 3 .* yerr)), showlegend=true, hoverinfo="skip", fill="toself", name="± 3STD")
        p.layout.xaxis.title = "Number of iterations"
        p.layout.yaxis.title = "Estimated scintillation index"
        p.layout.yaxis.range = [0, 2SI_gaussian]

        display(p)
    end

    if showplot
        # Plot SI distribution
        z = Adapt.adapt(Array, SI[:, :, 1])
        p = plot.heatmap(; x=fftshift(x), y=fftshift(y), z=fftshift(z), zmin=0, zmax=min(2mean(SI), 6))

        p.layout.showscale = true
        p.layout.xaxis.title = "x / r0"
        p.layout.yaxis.title = "y / r0"
        p.layout.yaxis.scaleanchor = "x"

        display(p)
    end

    return SI0
end


function coherence_length(fn; basedir=pwd(), debug=false, collimate=true)

    fn = joinpath(basedir, fn)

    # load data & metadata
    urx, metadata = h5open(fn, "r") do h5
        maybe_cuda(read(h5["RXFields"])), Dict(attrs(h5))
    end

    # coordinate axes
    Nx = metadata["Nx"]
    Ny = metadata["Ny"]
    x = fftfreq(Nx, last(metadata["Dx"]))
    y = fftfreq(Ny, last(metadata["Dy"]))
    # other parameters
    λ = metadata["lambda"]
    Dz = metadata["Dz"]
    r0 = metadata["r0eff"]

    # optionally, collimate before estimating params
    if collimate
        yt = transpose(y)
        r² = similar(urx, Nx, Ny)
        @. r² = x^2 + yt^2
        @. urx *= cispi(1 / (λ * Dz) * r²)
    end
    # fftshift axes
    x = fftshift(x)
    y = fftshift(y)

    # refer to approach by Schmidt
    mask = similar(urx, Nx, Ny)
    fill!(mask, 1)
    aperture!(mask, 1, Nx ÷ 4, centered=true)
    cf = real(cf_ft(fftshift(urx) .* mask; dims=1:2))
    CUDA.@allowscalar cf ./= first(cf)
    mcf = Adapt.adapt(Array, cf)[:, 1]

    # evaluate structure function on grid by integrating the appropriate PSD in "model"
    r = (0:Nx-1) * metadata["dx"][end]
    D_pw = num_sf(getfield(Main, Symbol(metadata["model"]))(1; κ0=2π / metadata["L0"], κl=3.3 / metadata["l0"], κm=5.92 / metadata["l0"]), r, 0.0, 1e-4)

    # analytical expression for plane wave MCF
    mcf_pw(x) = @. exp(-1 // 2 * 0.49 * x^(-5 // 3) * D_pw)
    # now, fit the estimated MCF to the analytical expression to estimate r0
    J(r0) = norm(mcf_pw(r0) .- mcf) # or coupling_efficiency?
    r0_est = Optim.minimizer(optimize(J, metadata["dx"][end], 10r0))

    if debug
        # theoretical mutual correlation function
        mcf_th = mcf_pw(r0)
        mcf_fit = mcf_pw(r0_est)

        @info "Estimated Fried parameter: " r0_est
        @info "Actual Fried parameter: " r0
        # MCF plots
        p = plot.scatter(; x=r ./ r0, y=mcf, name="Estimated MCF")
        p.scatter(; x=r ./ r0, y=mcf_fit, name="Fit")
        p.scatter(; x=r ./ r0, y=mcf_th, name="Theoretical MCF")

        p.layout.showscale = true
        p.layout.xaxis.title = "x / r0"

        display(p)
    end

    return r0_est
end


function beam_wander(fn; basedir=pwd(), debug=false)
    fn = joinpath(basedir, fn)

    # load data
    I = h5open(fn, "r") do h5
        abs2.(read(h5["RXFields"]))
    end

    In = I ./ sum(I; dims=1:2)

    # load metadata
    meta = attributes_from_dump(fn)

    # gather parameters
    Cn² = meta["Cn2"]
    Dz = meta["Dz"]
    w0 = meta["w0"]
    L0 = meta["L0"]
    κ0 = 2π / L0

    # gather grid
    x = fftfreq(meta["Nx"], meta["Nx"] * last(meta["dx"]))
    y = fftfreq(meta["Ny"], meta["Ny"] * last(meta["dy"]))
    yt = transpose(y)

    # first moment estimate of beam displacement in x/y direction, compare Schmidt 2022 p. 70 (p. 91 in pdf)
    cx = sum(x .* In; dims=1:2) |> vec
    cy = sum(yt .* In; dims=1:2) |> vec

    # identify hot spots and assign x/y positions
    hotspots = argmax(In; dims=1:2) |> vec
    xh = getindex(x, getindex.(hotspots, 1))
    yh = getindex(y, getindex.(hotspots, 2))
    
    # Beam-centroid wander: The mean centroid movement is zero so the second moment <βc²> corresponds to the variance of the centroid movement. 
    βc² = var(cx) + var(cy)

    # Total beam-wander variance: includes βc2 and the effect of more scales whose sizes are near that of the 1/e² beam diameter of the irradiance. Estimation based on ...
    # ... variance of the hot spot motion measured **from the boresight**
    σ²0 = var(xh) + var(yh)
    # ... variance of the hot spot motion as measured **relative to the centroid**
    σ²c = var(xh .- cx) + var(yh .- cy)
    # equation (3.12)
    rc² = σ²0 - σ²c

    # Theoretical reference values from Andrews
    rc²_ref = 2.42Cn² * Dz^3 / cbrt(w0) * (1 - (κ0^2 * w0^2 / (1 + κ0^2 * w0^2))^(1 / 6))
    βc²_ref = 0.56rc²_ref


    if debug
        @info "Estimated beam wander variance: " βc²
        @info "Theoretical beam wander variance: " βc²_ref

        @info "Estimated total beam wander variance: " rc²
        @info "Theoretical total beam wander variance: " rc²_ref

        βc²_evo = cumvar(cx) .+ cumvar(cy)
        σ0²_evo = cumvar(xh) .+ cumvar(yh)
        σc²_evo = cumvar(xh .- cx) + cumvar(yh .- cy)
        
        plt = plot.scatter(y=σ0²_evo .- σc²_evo, name="<rc²> - Estimate")
        plt.scatter(y=βc²_evo, name="<βc²> - Estimate")

        plt.scatter(y=fill(rc²_ref, length(σc²_evo)), name="<rc²> - Reference")
        plt.scatter(y=fill(βc²_ref, length(σc²_evo)), name="<βc²> - Reference")


        plt.layout.xaxis.title = "Iterations"
        plt.layout.yaxis.title = "Estimate"

        display(plt)
    end


    return rc²
end