include("analyze_turbulent_channel_model.jl")
include("analyze_coupling_data.jl")
include("repeated_wave_propagation.jl")

using CSV


function outer_product_across_dicts(dicts...)
    if isempty(dicts)
        return []
    end

    # Make sure the value arrays of each dict are of the same size
    for dict ∈ dicts
        valuelen = length.(values(dict))
        
        if !isone(length(unique(valuelen)))
            @info "Length of values:" Dict(k => length(v) for (k, v) ∈ dict)
            throw(ArgumentError("All values of an argument must have the same length."))
        end
    end

    # Extract keys and vectors from each dictionary
    dict_keys_vectors = [collect(pairs(dict)) for dict ∈ dicts]

    # Calculate the number of combinations
    dims = [length(first(kv)[2]) for kv ∈ dict_keys_vectors]

    # Initialize the result array
    result = Array{NamedTuple, length(dims)}(undef, dims...)

    # Generate all combinations of indices
    indices_combinations = Iterators.product(map(x -> 1:x, dims)...)

    # Populate the result array, use an OrderedDict to maintain the order!
    for indices ∈ indices_combinations
        current_dict = OrderedDict{Symbol, Any}()
        for (i, kv_pairs) ∈ enumerate(dict_keys_vectors)
            for (key, vector) ∈ kv_pairs
                current_dict[key] = vector[indices[i]]
            end
        end
        result[indices...] = NamedTuple(current_dict)
    end

    return result
end


function batch_repeated_wave_propagation(arg1=Dict(:Nz => 4:16, :dz => 1e3 ./ (3:15)), additional...; showprogress=true, samplingcheck=false, kwargs...)

    args = (arg1, additional...)

    dir = timeprefix(; fmt="YYYY-mm-dd", postfix="")

    # Build name of target directory
    for arg ∈ args
        for (k, v) ∈ pairs(arg)
            vmin, vmax = if eltype(v) <: Integer
                extrema(v)
            else
                round.(extrema(v), sigdigits=3)
            end
            
            dir *= "_$(k)_$(vmin)_$(vmax)"
        end
    end

    if isdir(dir)
        if !isempty(readdir(dir))
            @warn "Directory $dir already exists and is not empty! Do you want to write into this directory anyway? (y/N)"
            response = lowercase(strip(readline()))
            if response == "y"
                println("Continuing...")
            else
                println("Exiting...")
                return
            end
        end
    else
        mkdir(dir)
    end

    # Set up the array of NamedTuples for the iteration
    itr = outer_product_across_dicts(args...)
    Nv = length(itr)    

    p = Progress(length(itr); showspeed=true, enabled=showprogress)

    for (i, xarg) ∈ enumerate(itr)
        # we have to sort the outputs, why we prepend a number to the filenames
        repeated_wave_propagation(; xarg..., dir, fnprefix=lpad(i, ceil(Int, log10(Nv + 1)), "0") * "_", showprogress=false, samplingcheck, meta=xarg, kwargs...)
        sleep(0.1)
        next!(p, showvalues=[(string(k), v) for (k, v) ∈ pairs(xarg)])
    end

    # Save to CSV
    df = DataFrame(itr |> vec)
    open(joinpath(dir, "xvals.csv"), "w") do fid
        CSV.write(fid, df)
    end

    nothing
end

function batch_scintillation_index(dirname)

    # gather all hdf5 files in the given directory `dirname`
    files = readdir(dirname, join=true)
    h5files = filter(x -> splitext(x)[2] == ".hdf5", files)

    Nf = length(h5files)

    # gather xvals from corresponding csv if possible - otherwise generate xvals.csv and fill with 1:N
    xvals_fn = joinpath(dirname, "xvals.csv")
    !isfile(xvals_fn) && open(xvals_fn, "w") do fid
        CSV.write(fid, DataFrame(:file => range(1, Nf)))
    end
    df = CSV.read(xvals_fn, DataFrame)

    SI_est = zeros(Nf)
    SI_pw = zeros(Nf)
    SI_sw = zeros(Nf)
    SI_gaussian = zeros(Nf)

    p = Progress(Nf; showspeed=true)

    for (k, file) ∈ enumerate(h5files)
        SI_est[k] = scintillation_index(file)

        metadata = attributes_from_dump(file)

        # make sure the order the files are read matches the xvals in the DataFrame
        xname = first(names(df))
        if xname != "file"
            @assert df[k, xname] == metadata[string(xname)]
        end

        Cn² = metadata["Cn2"]
        λ = metadata["lambda"]
        Dz = metadata["Dz"]
        w0 = metadata["w0"]

        l0 = metadata["l0"]
        L0 = metadata["L0"]

        # the transverse grid resolution is inherently an inner scale of turbulence
        dxm = maximum(metadata["dx"])
        l0 = max(l0, dxm)

        SI_pw[k] = scintillation_index_pw(2π / λ, Dz, Cn², l0, L0)
        SI_sw[k] = scintillation_index_sw(2π / λ, Dz, Cn², l0, L0)
        SI_gaussian[k] = scintillation_index_gaussian(2π / λ, Dz, Cn², w0, l0, L0)

        next!(p, showvalues=[("file", file)])
    end

    # dump results
    df[!, "pw"] = SI_pw
    df[!, "sw"] = SI_sw
    df[!, "gaussian"] = SI_gaussian
    df[!, "estimate"] = SI_est

    open(joinpath(dirname, "scintillation_indices.csv"), "w") do fid
        CSV.write(fid, df)
    end

    nothing
end


function batch_coupling(; basedir=pwd())
    fns = readdir(basedir, join=true)
    filter!(contains(r"\d+\.hdf5"), fns)

    prg = Progress(length(fns), "Total progress", showspeed=true)
    for fn ∈ fns
        obtain_coupling_statistics(fn, optim_focus=true, dump=true, showprogress=false)
        next!(prg)
    end
end

function mean_couplingloss_vs_xval(xval="lambda"; basedir=pwd())
    fns = readdir(basedir, join=true)
    filter!(contains(r"coupling.hdf5"), fns)

    # for now, consider total coupling loss only
    df = DataFrame()

    for fn ∈ fns
        _, _, cL, meta = analyze_coupling_data(fn; Lrange=(0, 30))
        df[!, string(meta[xval])] = cL[!, :Total]
    end

    λ = parse.(Float64, names(df))
    Lm = vec(mean(Matrix(df), dims=1))
    
    plot.scatter(x=λ, y=Lm)
end